/*
 This project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Deviation is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Deviation.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libopencm3/stm32/f1/rcc.h>
#include <libopencm3/stm32/f1/gpio.h>
#include <libopencm3/stm32/nvic.h>
#include <libopencm3/stm32/exti.h>

#include <libopencm3/stm32/systick.h>
#include <libopencm3/stm32/timer.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/stm32/iwdg.h>
 
#include "common.h"
#include "interface.h"
#include "mixer.h"
#include "config/model.h"

//  #include "music.h"    // Debug,  for short to high-low , Trainer Rx(GPIO-PA10) test


/* ===ppmin.c===
(1) use TIM5 : set the unit same as "ppmout.c" for count the ppm-input signal, "uSecond (72MHz / 36) = 2MHz = 0.5uSecond"
(2) set GPIO PA10 input mode as external trigger 
(3) set GPIO PA10 into source(EXTI10) interrupt(NVIC_EXTI15_10_IRQ) to trigger-call "exti15_10_isr()" function
(4) transfer value for mixer.c   
	(4.1) "ppmin_num_channels"
	(4.2) "Channels[i]" or "raw[i+1]" : each channel value (volatile s16 Channels[NUM_OUT_CHANNELS];)
					Channels[i]
							= ((ppmChannels[i]*2MHz:uSecond - 1.5mSecond)/(1.0mSecond))*(CHAN_MAX_VALUE-CHAN_MIN_VALUE)
							= (ppmChannels[i]-3000)*10
	(4.3) "ppmSync"
 ===ppmin.c===*/

#ifndef EMULATOR

void ppmin_TIM5_Init()
{
	/* Enable TIM5 clock. */
	rcc_peripheral_enable_clock(&RCC_APB1ENR, RCC_APB1ENR_TIM5EN);
  
	/* No Enable TIM5 interrupt. */
	// nvic_enable_irq(NVIC_TIM5_IRQ);
	// nvic_set_priority(NVIC_TIM5_IRQ, 16); //High priority

	/* Reset TIM5 peripheral. */
	timer_disable_counter(TIM5);
	timer_reset(TIM5);

	/* Timer global mode:
	 * - No divider
	 * - Alignment edge
	 * - Direction up
	 */
	timer_set_mode(TIM5, TIM_CR1_CKD_CK_INT,
		       TIM_CR1_CMS_EDGE, TIM_CR1_DIR_UP);

   	/* Reset prescaler value.  timer updates each uSecond */
    timer_set_prescaler(TIM5, TIM5_prescaler);     // uSecond (72MHz / (35+1) = 2MHz = 0.5uSecond
    timer_set_period(TIM5, TIM5_period);    // 3300uSecond= 2MHz*6600times,  TIM5_prescaler=0.5uSecond

	/* Disable preload. */
	timer_disable_preload(TIM5);

	/* Continous mode. */
	timer_continuous_mode(TIM5);

    /* Disable outputs. */
    timer_disable_oc_output(TIM5, TIM_OC1);
    timer_disable_oc_output(TIM5, TIM_OC2);
    timer_disable_oc_output(TIM5, TIM_OC3);
    timer_disable_oc_output(TIM5, TIM_OC4);

	/* -- OC1 configuration -- */
	/* Configure global mode of line 1. */
	/* Enable CCP1 */
	timer_disable_oc_clear(TIM5, TIM_OC1);
	timer_disable_oc_preload(TIM5, TIM_OC1);
	timer_set_oc_slow_mode(TIM5, TIM_OC1);
	timer_set_oc_mode(TIM5, TIM_OC1, TIM_OCM_FROZEN);

	/* Enable commutation interrupt. */
	//  timer_enable_irq(TIM5, TIM_DIER_CC1IE);
	/* Disable CCP1 interrupt. */
    timer_disable_irq(TIM5, TIM_DIER_CC1IE);

	/* Counter enable. */
	timer_enable_counter(TIM5);
}

void ppmin_Init()
{
	ppmin_Stop_conflict();     // disable USART1 for GPIO PA9 & PA10 (Trainer Tx(PA9) & Rx(PA10))
	
	/* Enable GPIOA clock. */
	rcc_peripheral_enable_clock(&RCC_APB2ENR, RCC_APB2ENR_IOPAEN);

	/* Enable AFIO clock. */
	rcc_peripheral_enable_clock(&RCC_APB2ENR, RCC_APB2ENR_AFIOEN);

	/* Enable EXTI10 interrupt. */
	nvic_enable_irq(NVIC_EXTI15_10_IRQ);

	/* Set GPIO0 (in GPIO port A) to 'input float'. */
	gpio_set_mode(GPIOA, GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO10);
	/* Set GPIO-PA10  to Pull-Up. */
	//  gpio_set_mode(GPIOA, GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, GPIO10);
    //  gpio_set(GPIOA, GPIO10);   // setup GPIO-PA10 to Pull-Up
	//  gpio_clear(GPIOA, GPIO10);   // setup GPIO-PA10 to Pull-Down
	
	/* Configure the EXTI subsystem. */
	exti_select_source(EXTI10, GPIOA);
	//  exti_set_trigger(EXTI10, EXTI_TRIGGER_BOTH);     // Debug, for ppm-in:Rx(PA10) connect to ppm-out:Tx(PA9)
	exti_set_trigger(EXTI10, EXTI_TRIGGER_RISING);
	exti_enable_request(EXTI10);
}

/* ===get PPM===
(1) capture  ppmSync (for ppm-timing > MIN_PPMin_Sync : 3300uSecond)
(2) count channels and set to  "ppmin_num_channels"  (include Syn-singal)
(3) get  each channel value and set to  "Channels[i]" or "raw[i+1]"
			"Channels[i]" or "raw[i+1]"
					= ((ppmChannels[i]*2MHz:uSecond - 1.5mSecond)/(1.0mSecond))*(CHAN_MAX_VALUE-CHAN_MIN_VALUE)
					= (ppmChannels[i]-3000)*10
(4) continue count channels and compare  "num_channels",  if not equal => disconnect(no-Sync) and re-connect (re-Sync)
===get PPM===*/

volatile u8 ppmIN = 0;     //  ppm-in protocol status , 0: not use ppm-in protocol, 1: use ppm-in protocol , for mixer.c to check protocol status
volatile u8 ppmSync = 0;     //  the ppmSync for mixer.c,  0:ppm-Not-Sync , 1:ppm-Got-Sync
volatile u8 ppmin_num_channels;     //  the ppmin_num_channels for mixer.c 
volatile s16 ppmChannels[MAX_PPMin_Channels];    //  [0...ppmin_num_channels-1] for each channels width, [ppmin_num_channels] for sync-signal width
u8 k[4];
u8 j = 0;
u8 i = 0;
u16 t0 = 0;

void exti15_10_isr(void)
{
	u16 t1 = 0;
	u16 t = 0;
	
	t1 = timer_get_counter(TIM5);     // get the counter(TIM5) value	
	exti_reset_request(EXTI10);         //  reset(clean) the IRQ
	
	t = (t1>=t0) ? (t1-t0) : (65536+t1-t0);     // none-stop TIM5 counter, compute ppm-signal width (2MHz = 0.5uSecond)
	t0 = t1;
	
	//  gpio_toggle(GPIOA, GPIO9);     //  Debug, Trainer Tx:1/0 switch when ppm-in channel change
	
	if (!ppmSync) {      // ppm-in status : not Sync
		/*  (1) capture  pmSync (for ppm-timing > MIN_PPMin_Sync : 3300uSecond)  */
		if (t>MIN_PPMin_Sync) {    // ppm-input Sync-signal
			if (j<3) {       // set 3-times for count total channels number, k[0], k[1], k[2]
				j++;          // set for next count ppm-in total channels number
				k[j] = 0;    // initial ppm-in total channels number =0
			} else {         // accumulate 3-times total channels number k[0], k[1], k[2]
		/*  (2) count channels and set to  "ppmin_num_channels"  */
				j = 0;        // initial ppm-in Sync counter=0,  for re-connect Trainer (coach-trainee disconnect or noise)
				k[0] = 0;   // set ppm-in signal beginning k[0]=0, ignore the first count total channels number
				if (k[1]>1 && k[1]==k[2]) {     // compare total channels number k[1], k[2]
					ppmin_num_channels = k[1];        /*  transfer "ppmin_num_channels" to mixer.c   */
					ppmSync = 1;         /*  transfer "pmSync" to mixer.c   */
				}
			}
		} else {             // t<MIN_PPMin_Sync,  ppm-input each Channel-signal
			k[j]++;         // conut 3-times for total channels number, k[0], k[1], k[2]. ignore the first count total channels number k[0]
		}
	} else {                // ppm-in status : Sync, 
		/*  (3) get  each channel value and set to  "Channel[i]" ,  [0...ppmin_num_channels-1] for each Channel-signal */
		//   ppmChannels[i] = t;     // record ppm-signal width for each channels [0...ppmin_num_channels-1] , and Sync-signal width to [ppmin_num_channels]
		//   Channels[i] = (t-3000)*10;     /*   transfer channel value "Channels[i]" to mixer.c  , after mixer : MIXER_UpdateRawInputs()-5th step */
		raw[i+1] = (t-3000)*10;     /*   transfer channel value "raw[i]" to mixer.c  , before mixer : MIXER_UpdateRawInputs()-1st step */
		i++;		// set for next count  ppm-signal width
		/*  (4) continue count channels and compare  "num_channels",  if not equal => disconnect(no-Sync) and re-connect (re-Sync)  */
		if (t>MIN_PPMin_Sync) {     // Got the ppm-input Sync-signal 
			if ((i-1) != ppmin_num_channels) {          // Trainer disconnect (coach-trainee disconnect or noise)
				ppmSync = 0;                              // set ppm-in status to "Not Sync"
			}
		i = 0;      // initial counter for capture next period
		}
	}
}


void ppmin_Stop_conflict()
{
	UART_Stop();      // disable USART1 for GPIO PA9 & PA10 (Trainer Tx(PA9) & Rx(PA10))
	//  CLOCK_StopTimer();    // Debug, disable TIM4(ppm-out:Tx(PA9)), for ppm-in:Rx(PA10) connect to ppm-out:Tx(PA9)
}

void ppmin_Stop()
{
	nvic_disable_irq(NVIC_EXTI15_10_IRQ);
	exti_disable_request(EXTI10);
	ppmSync = 0;
}

void ppmin_Start()
{
	nvic_enable_irq(NVIC_EXTI15_10_IRQ);
	exti_enable_request(EXTI10);
	ppmSync = 0;
}
void ppmin_TIM5_Stop()

{
	timer_disable_counter(TIM5);
	ppmSync = 0;
}

void ppmin_TIM5_Start()
{
	timer_enable_counter(TIM5);
	ppmSync = 0;
}

#else     // #ifndef EMULATOR

/* EMULATOR function*/

#endif    // endif EMULATOR

